/*

th.c - Terminal Handler (UNIX Curses version)

Under Linux, we use ncurses. This doesn't provide termcap information.
It does support colour. We have no save/restore support.

Note that color pair 0 is not white-on-black as promised by 'man curs_color',
but is actually whatever the terminal believes is its 'normal' attribute.
This is what the color attribute is set to after endwin(). For an xterm,
this is often wrong, due to the terminfo entry having 'op=\E[30;47m', rather
than the 'op=\E[37;40m' you'd naturally expect.

Under AIX, we can use the older AIX Extended Curses, or the newer AIX Curses.
This does not provide any termcap information. It does support colour.

The default AIX terminfo database typically only has a monochrome entry
for vt100 and xterm. To get color vt100, I made vt100-color with base vt100
plus colfN/colbN entries stolen from other AIX terminfo entries.

On HP-UX we use Color Curses. There is no save/restore support.

To run in colour, this requires that there are suitable colour enabled terminfo
entries installed on the system. These are compiled with tic_colr rather than
the regular tic. I made vt100-color by taking the vt100 definition from my
Linux box, adding the 'setf', 'setb', 'setaf' and 'setab' definitions from the
linux terminfo definition, and then adding 'op=\E[m' and 'bce' entries to it.

Other UNIXes use vanilla curses, and expect termcap information to be
available. No save/restore support.

Changed so that \EOF and \EOH
  on AIX, assume hftterm, @F6 and @F8
  elsewhere, assume xterm, End and Home

*/

/*...sincludes:0:*/
#ifdef HP
#define	_XOPEN_SOURCE_EXTENDED
#undef __hpux
	/* Necessary to get the right colours */
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <setjmp.h>
#define	TH_C
#include "th.h"

/*...vth\46\h:0:*/

#ifdef AIX_EXTENDED_CURSES
#define	NO_TCAP
#include <cur00.h>	/* Base defns    */
#include <cur01.h>	/* Window stuff  */
#include <cur02.h>	/* Keycode stuff */
#include <cur05.h>	/* Panel stuff   */
#undef beep
#else
#include <curses.h>
#endif
/*...e*/
/*...sdefines:0:*/
#ifdef S390USS
#define	EBCDIC
#endif
/*...e*/
/*...svars:0:*/
int x_size, y_size;

static int csr_x, csr_y, csr_type = CSR_BLOCK;
static int cur_col;

static char blank_str[300];

#if defined(AIX) || defined(LINUX) || defined(HP) || defined(SUN) || defined(S390USS) || defined(CYGWIN) || defined(MACOSX) || defined(IPHONE)
static TH_BOOLEAN got_color = TH_FALSE;
#endif
/*...e*/

/*...sth_beep:0:*/
void th_beep(void)
	{
	/* not supported */
	}
/*...e*/
/*...sth_setcol:0:*/
/*
Must attempt to make use of the hilite and underline attributes etc., if
they are available. Note that they are mutually exclusive, since most
terminals (like VT100) do not give codes to change an attribute
independantly of the rest.
*/

#ifdef AIX_EXTENDED_CURSES
static int *fgs[] = { &F_BLACK, &F_BLUE, &F_GREEN, &F_CYAN,
		      &F_RED, &F_MAGENTA, &F_BROWN, &F_WHITE };
static int *bgs[] = { &B_BLACK, &B_BLUE, &B_GREEN, &B_CYAN,
		      &B_RED, &B_MAGENTA, &B_BROWN, &B_WHITE };
static int *bs[] = { &NORMAL, &UNDERSCORE, &STANDOUT, &BOLD };
#endif

void th_setcol(int col)
	{
	if ( col == cur_col )
		return;

#if defined(AIX_EXTENDED_CURSES)
	if ( got_color )
		colorout(*fgs[col&0x07] | *bgs[(col&0x70)>>4]);
	else
		colorout(*bs[(col&0x700)>>8]);
#elif defined(LINUX) || defined(HP) || defined(SUN) || defined(AIX_CURSES) || defined(S390USS) || defined(CYGWIN) || defined(MACOSX) || defined(IPHONE)
	if ( got_color )
		{
		int pair = 8*(7-(col&0x07))+((col&0x70)>>4);
		int attr = COLOR_PAIR(pair);
		if ( (col&0x07) == COL_YELLOW )
			attr |= A_BOLD;
		attrset(attr);
			/* Bug in ncurses curses.h means cannot put an
			   expression as argument to COLOR_PAIR macro.
			   Also, yellow foreground comes out brown,
			   unless you brighten it.
			   The 7-fg trick is because pair 0 is hard
			   wired to be white on black. */
		}
	else
		{
		if ( (cur_col&0x700) == COL_STAND )
			standend();
		if ( (col&0x700) == COL_STAND )
			standout();
		}
#else
	if ( (cur_col&0x700) == COL_STAND )
		standend();
	if ( (col&0x700) == COL_STAND )
		standout();
#endif

	cur_col = col;
	}
/*...e*/
/*...sth_setcsr:0:*/
void th_setcsr(int csr)
	{
	csr_type = csr;
	}
/*...e*/
/*...sth_cxy \43\ bits:0:*/
void th_cxy(int x, int y)
	{
	move(csr_y = y, csr_x = x);
	}

void th_cy(int y)
	{
	move(csr_y = y, csr_x = 0);
	}

void th_chome(void)
	{
	move(csr_y = 0, csr_x = 0);
	}
/*...e*/
/*...sth_cls:0:*/
void th_cls(void)
	{
	erase();
	move(csr_y = 0, csr_x = 0);
	}
/*...e*/
/*...sth_inslin:0:*/
void th_inslin(void)
	{
	insertln();
	move(csr_y, csr_x);
	}
/*...e*/
/*...sth_dellin:0:*/
void th_dellin(void)
	{
	deleteln();
	}
/*...e*/
/*...sth_deleol:0:*/
void th_deleol(void)
	{
	addstr(blank_str + csr_x);
	move(csr_y, csr_x);
	}
/*...e*/
#ifdef AIX_EXTENDED_CURSES
/*...sth_save:0:*/
void th_save(int x, int y, int n, void *buf)
	{
#if 0
	/* Try to use Extended Curses APIs to save characters and attributes on
	   the screen. Doesn't appear to work. Only characters get restored. */
	wchar_t *wbuf = (wchar_t *) buf;
	int i;
	for ( i = 0; i < n; i++ )
		wbuf[i] = mvinch(y, x+i);
	move(csr_y, csr_x);
#endif
	/* Take an inside peek in the WINDOW structure (yeuk!).
	   Copy both characters and attributes. */
	memcpy(buf, &(stdscr->_y[y][x]), n*sizeof(NLSCHAR));
	memcpy((NLSCHAR *) buf+n, &(stdscr->_a[y][x]), n*sizeof(ATTR));
	}
/*...e*/
/*...sth_restore:0:*/
void th_restore(int x, int y, int n, void *buf)
	{
#if 0
	/* Try to use Extended Curses APIs to save characters and attributes
	   on the screen. Doesn't appear to work. Only characters get restored.
	   Note also the A_ATTRIBUTES and A_CHARTEXT don't exist. */
	wchar_t *wbuf = (wchar_t *) buf;
	int i;
	for ( i = 0; i < n; i++ )
		{
		colorout(wbuf[i] & 0xfff00000);
		mvaddch(y, x+i, wbuf[i] & 0x000fffff);
		}
	move(csr_y, csr_x);
	th_setcol(cur_col);
#endif
	/* Nobble the WINDOW structure directly (yeuk!).
	   Copy both characters and attributes.
	   Need to be sure Extended Curses knows it must redraw the
	   screen at next refresh time. */
	memcpy(&(stdscr->_y[y][x]), buf, n*sizeof(NLSCHAR));
	memcpy(&(stdscr->_a[y][x]), (NLSCHAR *) buf+n, n*sizeof(ATTR));
	touchline(stdscr, y, x, n); /* Doesn't work! */
	touchwin(stdscr); /* So use a larger hammer */
	}
/*...e*/
#endif
/*...sth_isprint:0:*/
/* Technically, according to curses, only isprint() characters are printable.
   th_isprint returns those whose glyph can be displayed.
   isprint() returns TRUE for characters from ' ' to '~' on Linux.
   However, the input value is potentially negative.
   So we'll use a more restrictive (and non-locale sensitive) test. */

TH_BOOLEAN th_isprint(int c)
	{
#ifdef EBCDIC
	return isprint(c) && c != '\n' && c != '\t' && c != '\r' && c != '\b';
#else
	return c >= ' ' && c <= '~';
#endif
	}
/*...e*/
/*...sth_p_raw_chr:0:*/
/*
The th_p_raw_chr function below is only intended for use on normal ASCII (non
control) characters. Thus \n, \t etc. should not be sent using it (although it
would probably work).
*/

void th_p_raw_chr(int c)
	{
	addch(c);
	if ( (++csr_x) == x_size )
		{
		csr_x = 0; csr_y++;
		}
	}
/*...e*/
/*...sth_p_raw_str:0:*/
void th_p_raw_str(const char *s)
	{
#if defined(HP) || defined(SUN)
	addstr((char *) s); /* Missing const bug in curses.h */
#else
	addstr(s);
#endif
	if ( (csr_x += strlen(s)) >= x_size )
		{
		csr_y += (csr_x / x_size);
		csr_x = csr_x % x_size;
		}
	}
/*...e*/
/*...sth_p_chr:0:*/
/*
The th_p_raw_chr function and the th_p_raw_str function are for fast screen
update. As a result only normal ASCII characters should be sent using them.
For newline handling etc. use the 2 functions below.
*/

void th_p_chr(int c)
	{
	if ( c == '\n' )
		move(++csr_y, csr_x = 0);
	else if ( c == '\t' )
		do
			th_p_raw_chr(' ');
		while ( (csr_x & 7) != 0 );
	else if ( c == '\b' )
		{
		if ( csr_x > 0 )
			move(csr_y, --csr_x);
		}
	else
		th_p_raw_chr(c);
	}
/*...e*/
/*...sth_p_str:0:*/
void th_p_str(const char *s)
	{
	while ( *s )
		th_p_chr(*s++);
	}
/*...e*/
/*...sth_sync_screen:0:*/
void th_sync_screen(void)
	{
	if ( csr_type == CSR_OFF )
		move(y_size-1, x_size-1);
	refresh();
	if ( csr_type == CSR_OFF )
		move(csr_y, csr_x);
	}
/*...e*/
/*...skeyboard input:0:*/
/*
When reading a key from the keyboard, we must make special allowances
for the escape key. If escape is pressed, we must check for quickly
following keys, since this implies an arrow key has been pressed.
*/

#define	ESC 033

/*...slookup_key:0:*/
/* We have a hard-coded database of escape sequences.
   After searching this, we fall back to whatever curses may provide us with.
   We include VT100 codes as this is *very* common, and it is very common
   for terminals/termcap/terminfo entries to confuse cursor and application
   mode arrow keys - by storing both, we should work regardless.
   We know aixterm keys, and the hft differences (including OS/2 HFTTERM.EXE).
   We've got some sequences from the linux terminal too.
   We also know some native Sun and Xenix terminal escape codes.
   This section needs a restructuring.
   Many of the escape sequences listed apply to multiple terminal types. */

typedef struct { char *cap; int keycode; } AKC;

static AKC akc_tab[] =
	{
/*...svt100:8:*/
"\033[D",	_KE|K_LEFT,	/* VT100 Cursor mode arrows */
"\033[C",	_KE|K_RIGHT,
"\033[A",	_KE|K_UP,
"\033[B",	_KE|K_DOWN,

"\033OD",	_KE|K_LEFT,	/* VT100 Application mode arrows */
"\033OC",	_KE|K_RIGHT,
"\033OA",	_KE|K_UP,
"\033OB",	_KE|K_DOWN,

"\033OP",	_KE|K_F1,	/* VT100 Application mode f-keys */
"\033OQ",	_KE|K_F2,
"\033OR",	_KE|K_F3,
"\033OS",	_KE|K_F4,
/*...e*/
/*...sxterm:8:*/
"\033[11~",	_KE|K_F1, 
"\033[12~",	_KE|K_F2, 
"\033[13~",	_KE|K_F3, 
"\033[14~",	_KE|K_F4, 
"\033[15~",	_KE|K_F5, 
"\033[17~",	_KE|K_F6, 
"\033[18~",	_KE|K_F7, 
"\033[19~",	_KE|K_F8, 
"\033[20~",	_KE|K_F9, 
"\033[21~",	_KE|K_F10, 
"\033[23~",	_KE|K_F11, 
"\033[24~",	_KE|K_F12, 

#ifndef AIX

/* Issue is that these clash with alt number keys on hft terminal.
   Hack: Assume if not running on AIX, we're not using hft terminal. */

"\033O2P",	_KS|K_F1,
"\033O2Q",	_KS|K_F2,
"\033O2R",	_KS|K_F3,
"\033O2S",	_KS|K_F4,
"\033[15;2~",	_KS|K_F5,
"\033[17;2~",	_KS|K_F6,
"\033[18;2~",	_KS|K_F7,
"\033[19;2~",	_KS|K_F8,
"\033[20;2~",	_KS|K_F9,
"\033[21;2~",	_KS|K_F10,
"\033[23;2~",	_KS|K_F11,
"\033[24;2~",	_KS|K_F12,

"\033O5P",	_KC|K_F1,
"\033O5Q",	_KC|K_F2,
"\033O5R",	_KC|K_F3,
"\033O5S",	_KC|K_F4,
"\033[15;5~",	_KC|K_F5,
"\033[17;5~",	_KC|K_F6,
"\033[18;5~",	_KC|K_F7,
"\033[19;5~",	_KC|K_F8,
"\033[20;5~",	_KC|K_F9,
"\033[21;5~",	_KC|K_F10,
"\033[23;5~",	_KC|K_F11,
"\033[24;5~",	_KC|K_F12,

"\033O3P",	_KA|K_F1,
"\033O3Q",	_KA|K_F2,
"\033O3R",	_KA|K_F3,
"\033O3S",	_KA|K_F4,
"\033[15;3~",	_KA|K_F5,
"\033[17;3~",	_KA|K_F6,
"\033[18;3~",	_KA|K_F7,
"\033[19;3~",	_KA|K_F8,
"\033[20;3~",	_KA|K_F9,
"\033[21;3~",	_KA|K_F10,
"\033[23;3~",	_KA|K_F11,
"\033[24;3~",	_KA|K_F12,

/* Issue is that these clash with @F6 and @F8 on hft terminal.
   Hack: Assume if not running on AIX, we're not using hft terminal. */

"\033OF",	_KE|K_END,
"\033OH",	_KE|K_HOME,

#endif

"\033Op",	_KE|K_INSERT,
"\033On",	_KE|K_DELETE,
"\033[H",	_KE|K_HOME,
"\033[F",	_KE|K_END,
"\033[5~",	_KE|K_PGUP,
"\033[6~",	_KE|K_PGDN,
/*...e*/
/*...srxvt:8:*/
/* rxvt also generates a couple of extras over standard xterm.
   ~Fn is like pressing F(n+10), so ~F1 is indistinguishable from F11,
   so ignore these. Same for ~^Fn being like ^F(n+10). */

"\033[2~",	_KE|K_INSERT,
"\033[3~",	_KE|K_DELETE,
"\033[7~",	_KE|K_HOME,
"\033[8~",	_KE|K_END,

"\033[2^",	K_CTRL_INSERT,
"\033[3^",	K_CTRL_DELETE,
"\033[7^",	K_CTRL_HOME,
"\033[8^",	K_CTRL_END,
"\033[5^",	K_CTRL_PGUP,
"\033[6^",	K_CTRL_PGDN,

"\033[11^",	K_CTRL_F1, 
"\033[12^",	K_CTRL_F2, 
"\033[13^",	K_CTRL_F3, 
"\033[14^",	K_CTRL_F4, 
"\033[15^",	K_CTRL_F5, 
"\033[17^",	K_CTRL_F6, 
"\033[18^",	K_CTRL_F7, 
"\033[19^",	K_CTRL_F8, 
"\033[20^",	K_CTRL_F9, 
"\033[21^",	K_CTRL_F10, 
"\033[23^",	K_CTRL_F11, 
"\033[24^",	K_CTRL_F12, 
/*...e*/
/*...sesc\45\x is alt\43\x:8:*/
/* linux and rxvt terminals generate esc-x when alt+x pressed.
   This is great for everyone, especially as not all terminals have alt. */

"\033a",	K_ALT_A,
"\033b",	K_ALT_B,
"\033c",	K_ALT_C,
"\033d",	K_ALT_D,
"\033e",	K_ALT_E,
"\033f",	K_ALT_F,
"\033g",	K_ALT_G,
"\033h",	K_ALT_H,
"\033i",	K_ALT_I,
"\033j",	K_ALT_J,
"\033k",	K_ALT_K,
"\033l",	K_ALT_L,
"\033m",	K_ALT_M,
"\033n",	K_ALT_N,
"\033o",	K_ALT_O,
"\033p",	K_ALT_P,
"\033q",	K_ALT_Q,
"\033r",	K_ALT_R,
"\033s",	K_ALT_S,
"\033t",	K_ALT_T,
"\033u",	K_ALT_U,
"\033v",	K_ALT_V,
"\033w",	K_ALT_W,
"\033x",	K_ALT_X,
"\033y",	K_ALT_Y,
"\033z",	K_ALT_Z,

"\0330",	K_ALT_0,
"\0331",	K_ALT_1,
"\0332",	K_ALT_2,
"\0333",	K_ALT_3,
"\0334",	K_ALT_4,
"\0335",	K_ALT_5,
"\0336",	K_ALT_6,
"\0337",	K_ALT_7,
"\0338",	K_ALT_8,
"\0339",	K_ALT_9,
/*...e*/
/*...slinux:8:*/
/* Basically like xterm and rxvt, but with a few variations. */

"\033[[A",	_KE|K_F1, 
"\033[[B",	_KE|K_F2, 
"\033[[C",	_KE|K_F3, 
"\033[[D",	_KE|K_F4, 
"\033[[E",	_KE|K_F5, 

"\033[1~",	_KE|K_HOME,
"\033[4~",	_KE|K_END,
/*...e*/
/*...saixterm:8:*/
/* Like vt100 and xterm somewhat, but with */

"\033[001q",	_KE|K_F1, 
"\033[002q",	_KE|K_F2, 
"\033[003q",	_KE|K_F3, 
"\033[004q",	_KE|K_F4, 
"\033[005q",	_KE|K_F5, 
"\033[006q",	_KE|K_F6, 
"\033[007q",	_KE|K_F7, 
"\033[008q",	_KE|K_F8, 
"\033[009q",	_KE|K_F9, 
"\033[010q",	_KE|K_F10, 
"\033[011q",	_KE|K_F11, 
"\033[012q",	_KE|K_F12, 

"\033[013q",	K_SHIFT_F1, 
"\033[014q",	K_SHIFT_F2, 
"\033[015q",	K_SHIFT_F3, 
"\033[016q",	K_SHIFT_F4, 
"\033[017q",	K_SHIFT_F5, 
"\033[018q",	K_SHIFT_F6, 
"\033[019q",	K_SHIFT_F7, 
"\033[020q",	K_SHIFT_F8, 
"\033[021q",	K_SHIFT_F9, 
"\033[022q",	K_SHIFT_F10, 
"\033[023q",	K_SHIFT_F11, 
"\033[024q",	K_SHIFT_F12, 

"\033[025q",	K_CTRL_F1, 
"\033[026q",	K_CTRL_F2, 
"\033[027q",	K_CTRL_F3, 
"\033[028q",	K_CTRL_F4, 
"\033[029q",	K_CTRL_F5, 
"\033[030q",	K_CTRL_F6, 
"\033[031q",	K_CTRL_F7, 
"\033[032q",	K_CTRL_F8, 
"\033[033q",	K_CTRL_F9, 
"\033[034q",	K_CTRL_F10, 
"\033[035q",	K_CTRL_F11, 
"\033[036q",	K_CTRL_F12, 

"\033[037q",	K_ALT_F1, 
"\033[038q",	K_ALT_F2, 
"\033[039q",	K_ALT_F3, 
"\033[040q",	K_ALT_F4, 
"\033[041q",	K_ALT_F5, 
"\033[042q",	K_ALT_F6, 
"\033[043q",	K_ALT_F7, 
"\033[044q",	K_ALT_F8, 
"\033[045q",	K_ALT_F9, 
"\033[046q",	K_ALT_F10, 
"\033[047q",	K_ALT_F11, 
"\033[048q",	K_ALT_F12, 

"\033[058q",	K_ALT_1, 
"\033[059q",	K_ALT_2, 
"\033[060q",	K_ALT_3, 
"\033[061q",	K_ALT_4, 
"\033[062q",	K_ALT_5, 
"\033[063q",	K_ALT_6, 
"\033[064q",	K_ALT_7, 
"\033[065q",	K_ALT_8, 
"\033[066q",	K_ALT_9, 
"\033[067q",	K_ALT_0, 

"\033[074q",	K_ALT_Q, 
"\033[075q",	K_ALT_W, 
"\033[076q",	K_ALT_E, 
"\033[077q",	K_ALT_R, 
"\033[078q",	K_ALT_T, 
"\033[079q",	K_ALT_Y, 
"\033[080q",	K_ALT_U, 
"\033[081q",	K_ALT_I, 
"\033[082q",	K_ALT_O, 
"\033[083q",	K_ALT_P, 

"\033[087q",	K_ALT_A, 
"\033[088q",	K_ALT_S, 
"\033[089q",	K_ALT_D, 
"\033[090q",	K_ALT_F, 
"\033[091q",	K_ALT_G, 
"\033[092q",	K_ALT_H, 
"\033[093q",	K_ALT_J, 
"\033[094q",	K_ALT_K, 
"\033[095q",	K_ALT_L, 

"\033[101q",	K_ALT_Z, 
"\033[102q",	K_ALT_X, 
"\033[103q",	K_ALT_C, 
"\033[104q",	K_ALT_V, 
"\033[105q",	K_ALT_B, 
"\033[106q",	K_ALT_N, 
"\033[107q",	K_ALT_M, 

"\033[146q",	_KE|K_END,
"\033[150q",	_KE|K_PGUP,
"\033[154q",	_KE|K_PGDN,
"\033[139q",	_KE|K_INSERT,		/* Ins and ~Ins */
"\033[P",	_KE|K_DELETE,		/* Del and ~Del */

"\033[158q",	K_SHIFT_LEFT,
"\033[167q",	K_SHIFT_RIGHT,
"\033[161q",	K_SHIFT_UP,
"\033[164q",	K_SHIFT_DOWN,
"\033[143q",	K_SHIFT_HOME,
"\033[147q",	K_SHIFT_END,
"\033[155q",	K_SHIFT_PGDN,
"\033[151q",	K_SHIFT_PGUP,

"\033[159q",	K_CTRL_LEFT,
"\033[168q",	K_CTRL_RIGHT,
"\033[162q",	K_CTRL_UP,
"\033[165q",	K_CTRL_DOWN,
"\033[144q",	K_CTRL_HOME,
"\033[148q",	K_CTRL_END,	
"\033[156q",	K_CTRL_PGDN,
"\033[152q",	K_CTRL_PGUP,
"\033[140q",	K_CTRL_INSERT,
"\033[142q",	K_CTRL_DELETE,
"\033[200q",	K_CTRL_G_PLUS,
"\033[198q",	K_CTRL_G_MINUS,
"\033[187q",	K_CTRL_G_STAR,
"\033[179q",	K_CTRL_G_SLASH,

"\033[160q",	K_ALT_LEFT,
"\033[169q",	K_ALT_RIGHT,
"\033[163q",	K_ALT_UP,
"\033[166q",	K_ALT_DOWN,
"\033[145q",	K_ALT_HOME,
"\033[149q",	K_ALT_END,
"\033[153q",	K_ALT_PGUP,
"\033[157q",	K_ALT_PGDN,
"\033[141q",	K_ALT_INSERT,
"\033[M",	K_ALT_DELETE,
"\033[201q",	K_ALT_G_PLUS,
"\033[199q",	K_ALT_G_MINUS,
"\033[188q",	K_ALT_G_STAR,
"\033[180q",	K_ALT_G_SLASH,
/*...e*/
/*...shft differences from aixterm:8:*/
"\033Oa",	K_CTRL_F1, 
"\033Ob",	K_CTRL_F2, 
"\033Oc",	K_CTRL_F3, 
"\033Od",	K_CTRL_F4, 
"\033Oe",	K_CTRL_F5, 
"\033Of",	K_CTRL_F6, 
"\033Og",	K_CTRL_F7, 
"\033Oh",	K_CTRL_F8, 
"\033Oi",	K_CTRL_F9, 
"\033Oj",	K_CTRL_F10, 

"\033OK",	K_ALT_F1, 
"\033OL",	K_ALT_F2, 
"\033OO",	K_ALT_F3, 
"\033ON",	K_ALT_F4, 
"\033OE",	K_ALT_F5, 
"\033OG",	K_ALT_F7, 
"\033OI",	K_ALT_F9, 
"\033OJ",	K_ALT_F10, 

#ifdef AIX

/* Issue is that these clash with Home and End on Gnome xterm terminal.
   Hack: If running on AIX, give hft definition preference. */

"\033OF",	K_ALT_F6,
"\033OH",	K_ALT_F8,

/* Issue is that these clash with function keys on xterm.
   Hack: If running on AIX, give hft definition preference. */

"\033O1",	K_ALT_1, 
"\033O2",	K_ALT_2, 
"\033O3",	K_ALT_3, 
"\033O4",	K_ALT_4, 
"\033O5",	K_ALT_5, 
"\033O6",	K_ALT_6, 
"\033O7",	K_ALT_7, 
"\033O8",	K_ALT_8, 
"\033O9",	K_ALT_9, 
"\033O0",	K_ALT_0, 

#endif
/*...e*/
/*...sSun:8:*/
/* Escape sequences for the L1-L10 keystrokes should be mapped to the
   K_ALT_F(n) keycodes here. However, I've lost the list over the years.
   The following mapping should also be encoded here :-
     R7 ->Home   R8->Up      R9->PgUp
     R10->Left               R12->Right
     R13->End    R14->Down   R15->PgDn
   Other Rn right-hand function keys are not supported.
   To be added when I next get access to a suitably configured Sun system. */
/*...e*/
/*...sXenix:8:*/
/* Escape sequences for the F1-F10, ~F1-10 and ^F1-10 keystrokes should
   be encoded here, although I've lost the list over the years.
   Also, ~^F1-F10 should be mapped to @F1-10.
   Entries to be added when I next get access to a suitable Xenix system. */
/*...e*/
	};

#define	AKC_ENTRIES (sizeof(akc_tab)/sizeof(akc_tab[0]))

#if !defined(AIX) && !defined(LINUX) && !defined(HP) && !defined(SUN) && !defined(S390USS) && !defined(CYGWIN) && !defined(MACOSX) && !defined(IPHONE)

typedef struct { char **cap; int keycode; } KC;

static KC kc_tab[] =
	{
	&KH,		_KE|K_HOME,
	&KL,		_KE|K_LEFT,
	&KR,		_KE|K_RIGHT,
	&KU,		_KE|K_UP,
	&KD,		_KE|K_DOWN,

	&K1,		_KE|K_F1,
	&K2,		_KE|K_F2,
	&K3,		_KE|K_F3,
	&K4,		_KE|K_F4,
	&K5,		_KE|K_F5,
	&K6,		_KE|K_F6,
	&K7,		_KE|K_F7,
	&K8,		_KE|K_F8,
	&K9,		_KE|K_F9,
	&K0,		_KE|K_F10,
	};

#define	KC_ENTRIES (sizeof(kc_tab)/sizeof(kc_tab[0]))

#endif

static int lookup_key(const char *s)
	{
	int i;
	for ( i = 0; i < AKC_ENTRIES; i++ )
		if ( !strcmp(akc_tab[i].cap, s) )
			return akc_tab[i].keycode&~_KE;
#if !defined(AIX) && !defined(LINUX) && !defined(HP) && !defined(SUN) && !defined(S390USS) && !defined(CYGWIN) && !defined(MACOSX) && !defined(IPHONE)
	for ( i = 0; i < KC_ENTRIES; i++ )
		if ( *(kc_tab[i].cap) != NULL && !strcmp(*(kc_tab[i].cap), s) )
			return kc_tab[i].keycode&~_KE;
#endif
	if ( s[0] == ESC || s[1] == ESC )
		/* Make another pass of the tables, hoping to pick
		   up upon Esc-Fn being mapped to Alt+Fn etc.. */
		{
		for ( i = 0; i < AKC_ENTRIES; i++ )
			if ( (akc_tab[i].keycode & _KE) != 0 &&
			     !strcmp(akc_tab[i].cap, s+1) )
				return (akc_tab[i].keycode & ~_KE) | _KA;
#if !defined(AIX) && !defined(LINUX) && !defined(HP) && !defined(SUN) && !defined(S390USS) && !defined(CYGWIN) && !defined(MACOSX) && !defined(IPHONE)
		for ( i = 0; i < KC_ENTRIES; i++ )
			if ( (kc_tab[i].keycode & _KE) != 0 &&
			     *(kc_tab[i].cap) != NULL && !strcmp(*(kc_tab[i].cap), s+1) )
				return (kc_tab[i].keycode & ~_KE) | _KA;
#endif
		}
	return '\0';
	}
/*...e*/

#define	N_AHEAD	30
static unsigned char ahead[N_AHEAD+1];
static int ahead_used = 0;

static TH_BOOLEAN had_key_resize = FALSE;

/*...sgetch_non_error:0:*/
static int getch_non_error(void)
	{
	int k;
	while ( (k = getch()) == ERR )
		;
	return k;
	}
/*...e*/
/*...sdecode_ahead:0:*/
/* There is guaranteed to be at least one character in the 'ahead' buffer on
   entry to this routine. It returns the decoded key. Characters may remain
   in the 'ahead' buffer afterwards. If the characters in the ahead buffer
   form an unrecognised escape sequence, it will try to accumulate more into
   the buffer. It can give up if 1s expires, or the maximum escape sequence
   length is exceeded.

   Note that Linux 2.2.x seems to behave differently than Linux 2.0.x and
   all other UNIXes I've tried in that I must use sigsetjmp(sjbuf,1) so as
   to be sure the signal masking state is restored. It looks as if the first
   time the SIGALRM is delivered to onalarm, because onalarm never returns
   the SIGALRM remains masked forever, despite calls to alarm(). In this code
   I assume that if I'm compiling for Linux, its 2.2.x or later. */

#if defined(LINUX) || defined(S390USS) || defined(CYGWIN) || defined(MACOSX) || defined(IPHONE)
static sigjmp_buf sjbuf;
#else
static jmp_buf sjbuf;
#endif

/*...sonalarm:0:*/
#if defined(AIX) || defined(LINUX) || defined(HP) || defined(SUN) || defined(S390USS) || defined(CYGWIN) || defined(MACOSX) || defined(IPHONE)
static void onalarm(int x)
#else
static int onalarm()
#endif
	{
#if defined(LINUX) || defined(S390USS) || defined(CYGWIN) || defined(MACOSX) || defined(IPHONE)
	siglongjmp(sjbuf, 1);
#else
	longjmp(sjbuf, 1);
#endif
#if !defined(AIX) && !defined(LINUX) && !defined(HP) && !defined(SUN) && !defined(S390USS) && !defined(CYGWIN) && !defined(MACOSX) && !defined(IPHONE)
	return 0;
#endif
	}
/*...e*/

static int decode_ahead(void)
	{
	int k;
#ifdef HP
	signal(_SIGALRM, onalarm);
#else
	signal(SIGALRM, onalarm);
#endif
	alarm(1);
#if defined(LINUX) || defined(S390USS) || defined(CYGWIN) || defined(MACOSX) || defined(IPHONE)
	if ( sigsetjmp(sjbuf, 1) == 0 )
#else
	if ( setjmp(sjbuf) == 0 )
#endif
		for ( ;; )
			{
			if ( (k = ahead[0]) != ESC )
				{
				alarm(0);
				memmove(ahead, ahead+1, ahead_used--);
				break;
				}
			if ( (k = lookup_key((const char *) ahead)) != '\0' )
				{
				alarm(0);
				ahead_used = 0;
				break;
				}
			if ( ahead_used == N_AHEAD )
				{
				alarm(0);
				k = ahead[0];
				memmove(ahead, ahead+1, ahead_used--);
				break;
				}
			k = getch_non_error();
#ifdef KEY_RESIZE
			if ( k == KEY_RESIZE )
				{
				k = K_RESIZE;
				ahead_used = 0;
				break;
				}
#endif
			ahead[ahead_used+1] = '\0';
			ahead[ahead_used++] = (unsigned char) k;
			}
	else
		{
#ifdef IPHONE
		/* On iPod Touch 1.1.4, sigsetjmp and siglongjmp do not
		   save and/or restore the signal mask as expected/required.
		   manually clear the alarm mask */
		sigset_t t;
		sigemptyset(&t);
		sigaddset(&t, SIGALRM);
		sigprocmask(SIG_UNBLOCK, &t, NULL);
#endif
		alarm(0);
		k = ahead[0];
		memmove(ahead, ahead+1, ahead_used--);
		}
#ifdef EBCDIC
	/* EBCDIC code points can legimately be above 0x7f */
#else
	if ( (k & 0xff80) == 0x80 )
		/* Its a 'meta' key - ie: Alt */
		{
		int c = (k&0x7f);
		if ( c >= 'A' && c <= 'Z' )
			k = K_AC(c);
		else if ( c >= 'a' && c <= 'z' )
			k = K_AC(c-'a'+'A');
		else if ( c >= '0' && c <= '9' )
			k = K_AN(c-'0');
		else
			k = -1;
		}
#endif
	return k;
	}
/*...e*/
/*...skey_read:0:*/
static int key_read(void)
	{
	if ( ahead_used == 0 )
		{
		int k = getch_non_error();
#ifdef KEY_RESIZE
		if ( k == KEY_RESIZE )
			return K_RESIZE;
#endif
		ahead[ahead_used++] = (unsigned char) k;
		ahead[ahead_used  ] = '\0';
		}
	return decode_ahead();
	}
/*...e*/

static TH_BOOLEAN got_key = TH_FALSE;
static int key_got;

/*...sth_key_read:0:*/
int th_key_read(void)
	{
	int k;
	if ( csr_type == CSR_OFF )
		move(y_size-1, x_size-1);
	refresh();
	if ( got_key )
		{
		got_key = TH_FALSE;
		k = key_got;
		}
	else
		k = key_read();
	if ( csr_type == CSR_OFF )
		move(csr_y, csr_x);
	return k;
	}
/*...e*/
/*...sth_key_status:0:*/
#if defined(AIX) || defined(LINUX) || defined(HP) || defined(SUN) || defined(S390USS) || defined(CYGWIN) || defined(MACOSX) || defined(IPHONE)
int th_key_status(void)
	{
	int k;
	if ( got_key )
		return key_got;
#ifdef AIX_EXTENDED_CURSES
	nodelay(TH_TRUE);
	k = getch();
	nodelay(TH_FALSE);
	if ( k == KEY_NOKEY )
		return 0;
#else
	nodelay(stdscr, TH_TRUE);
	k = getch();
	nodelay(stdscr, TH_FALSE);
	if ( k == ERR )
		return 0;
#endif
	ahead[ahead_used++] = k;
	ahead[ahead_used  ] = '\0';
	key_got = decode_ahead();
	got_key = TH_TRUE;
	return key_got;
	}
#endif
/*...e*/
/*...sth_key_break:0:*/
static TH_BOOLEAN intr_occurred = TH_FALSE;

TH_BOOLEAN th_key_break(void)
	{
	TH_BOOLEAN break_occurred = intr_occurred;
	intr_occurred = TH_FALSE;
	return break_occurred;
	}
/*...e*/

/*...sonintr:0:*/
#if defined(AIX) || defined(LINUX) || defined(HP) || defined(SUN) || defined(S390USS) || defined(CYGWIN) || defined(MACOSX) || defined(IPHONE)
void onintr(int x)
#else
void onintr()
#endif
	{
	intr_occurred = TH_TRUE;
	signal(SIGINT, onintr);
	}
/*...e*/
/*...e*/
/*...sth_init and th_deinit:0:*/
static TH_BOOLEAN th_inited = TH_FALSE;

TH_BOOLEAN th_init(int mode, int x_size_reqd, int y_size_reqd)
	{
#ifndef AIX_CURSES
	if ( freopen("/dev/tty", "w", stdout) == NULL ||
	     freopen("/dev/tty", "r", stdin ) == NULL )
		return TH_FALSE;
#endif

	initscr();
		/* Bug: COLS and LINES are not to be trusted if window has
		   been resized since process started. */

	if ( (x_size_reqd != 0 && x_size_reqd != COLS ) ||
	     (y_size_reqd != 0 && y_size_reqd != LINES) )
		{
		endwin();
		return TH_FALSE;
		}

#if defined(LINUX) || defined(HP) || defined(SUN) || defined(AIX_CURSES) || defined(S390USS) || defined(CYGWIN) || defined(MACOSX) || defined(IPHONE)
	/* Calling start_color() on AIX causes a segmentation violation.
	   Known to do this on AIX 4.1.x and AIX 4.2.x.
	   This is true regardless of whether the terminal supports color. */
	start_color();
#endif
#if defined(AIX) || defined(LINUX) || defined(HP) || defined(SUN) || defined(S390USS) || defined(CYGWIN) || defined(MACOSX) || defined(IPHONE)
	switch ( mode )
		{
		case MODE_COL:
			got_color = TH_TRUE;
			break;
		case MODE_NO_COL:
			got_color = TH_FALSE;
			break;
		default:
#ifdef AIX_EXTENDED_CURSES
			/* AIX Extended Curses version of has_colors()
			   didn't work properly when I tried it.
			   It always returns TH_TRUE!
			   Known to do this on AIX 4.1.x and AIX 4.2.x.
			   Hardcode a list of known color terminals. */
			{
			const char *p;
			if ( (p = getenv("TERM")) != NULL )
				if ( !strcmp(p, "aixterm"    ) ||
				     !strcmp(p, "hft"        ) ||
				     !strcmp(p, "linux"      ) ||
				     !strcmp(p, "LINUX"      ) ||
				     !strcmp(p, "vt100-color") ||
				     !strcmp(p, "xterm-color") ||
				     !strcmp(p, "rxvt"       ) )
					got_color = TH_TRUE;
			}
#else
			got_color = has_colors();
#endif
			break;
		}
#endif

#if defined(LINUX) || defined(HP) || defined(SUN) || defined(AIX_CURSES) || defined(S390USS) || defined(CYGWIN) || defined(MACOSX) || defined(IPHONE)
	if ( got_color )
		{
		static int cols[] = {
			COLOR_BLACK, COLOR_BLUE, COLOR_GREEN, COLOR_CYAN,
			COLOR_RED, COLOR_MAGENTA, COLOR_YELLOW, COLOR_WHITE };
		int f, b;
		for ( f = 0; f < 8; f++ )
			for ( b = 0; b < 8; b++ )
				init_pair((7-f)*8+b, cols[f], cols[b]);
		}
#endif

	noecho();
#ifdef AIX_EXTENDED_CURSES
	crmode();
#else
	cbreak();
#endif
	nonl();
	raw();

#if !defined(AIX) && !defined(LINUX) && !defined(HP) && !defined(SUN) && !defined(S390USS) && !defined(CYGWIN) && !defined(MACOSX) && !defined(IPHONE)
	if ( KS != NULL ) fputs(KS, stdout);
#endif

	x_size = COLS;
	y_size = LINES;

	memset(blank_str, ' ', x_size);
	blank_str[x_size] = '\0';

#ifdef AIX_EXTENDED_CURSES
	wcolorout(curscr, F_BLACK|B_WHITE); werase(curscr);
#endif
	th_setcol(COL(COL_WHITE,COL_BLACK,COL_NORM));
	th_setcsr(CSR_BLOCK);
	th_cls();
	refresh();

	if ( signal(SIGINT, SIG_IGN) != SIG_IGN )
		signal(SIGINT, onintr);

	return th_inited = TH_TRUE;
	}

void th_deinit(void)
	{
	if ( !th_inited )
		return;

	th_setcol(COL(COL_WHITE,COL_BLACK,COL_NORM));
	th_setcsr(CSR_BLOCK);
	th_cls();
	refresh();

#if !defined(AIX) && !defined(LINUX) && !defined(HP) && !defined(SUN) && !defined(S390USS) && !defined(CYGWIN) && !defined(MACOSX) && !defined(IPHONE)
	if ( KE != NULL ) fputs(KE, stdout);
#endif

	noraw();
	nl();
	echo();
#ifdef AIX_EXTENDED_CURSES
	nocrmode();
#else
	nocbreak();
#endif
	endwin();
	}
/*...e*/
