/*

th.h - Interface to Terminal Handler (UNIX curses version)

This code can use the basic curses support. This is rather limited.

The set of keystrokes support depends most upon the terminal type being used.
So this file provides access to all common known keystrokes. Different
terminal types provide different subsets. Listed next to each key K_ value,
is the terminals which can reasonably be expected to produce it.

The default Xenix terminal generates keycodes for Ctrl+Shift+Fn key
combinations. These are mapped to Alt+Fn keystrokes.

The default SunOS terminal generates keycodes for its L1-L10 left function
pad keys. These are mapped to Alt+Fn keystrokes. Parts of the R1-R15 keypad
are mapped into the PC style keypad.

*/

#ifndef TH_H
#define	TH_H

typedef int TH_BOOLEAN;
#define	TH_TRUE  1
#define	TH_FALSE 0

#define	CSR_BLOCK	0		/* Block cursor                      */
#define	CSR_ULINE	1		/* Underline cursor if exist         */
#define CSR_ALT		2		/* Alternate cursor if exist         */
#define	CSR_OFF		3		/* Make cursor invisible             */

#define	COL_BLACK	0
#define	COL_BLUE	1
#define	COL_GREEN	2
#define	COL_CYAN	3
#define	COL_RED		4
#define	COL_MAGENTA	5
#define	COL_YELLOW	6
#define	COL_WHITE	7
#define	COL_NORM	0x000		/* Normal attributes                 */
#define	COL_ULINE	0x100		/* Use underline attribute           */
#define	COL_STAND	0x200		/* Use standout attribute            */
#define	COL_ALT		0x300		/* Alternate character set           */

#define	COL(fg,bg,b)	((fg)+((bg)<<4)+(b))

#define	MODE_NORMAL	0x0000		/* Auto-detects colour support       */
#define	MODE_COL	0x0001		/* Forces use of colour              */
#define	MODE_NO_COL	0x0002		/* Forces no use of colour           */

#define	_KS		0x1000
#define	_KC		0x2000
#define	_KA		0x4000
#define	_KE		0x8000

#define	K_BS		0x0008		/* Any terminal */
#define	K_TAB		0x0009		/* Any terminal */
#define	K_LF		0x000a		/* Any terminal */
#define	K_CR		0x000d		/* Any terminal */
#define	K_ESC		0x001b		/* Any terminal */
#define	K_CTRL(c)	((c)&0x1f)	/* Any terminal */
#define	K_DEL		0x007f		/* Any terminal */

#define	K_LEFT		0x0100		/* Any terminal */
#define	K_RIGHT		0x0101		/* Any terminal */
#define	K_UP		0x0102		/* Any terminal */
#define	K_DOWN		0x0103		/* Any terminal */
#define	K_HOME		0x0104		/* Any terminal */
#define	K_END		0x0105		/* aixterm,hft,linux */
#define	K_PGUP		0x0106		/* xterm,aixterm,hft,linux */
#define	K_PGDN		0x0107		/* xterm,aixterm,hft,linux */
#define	K_INSERT	0x0108		/* xterm,aixterm,hft,linux */
#define	K_DELETE	0x0109		/* xterm,aixterm,hft,linux */
#define	K_G_PLUS	0x010a		/* none */
#define	K_G_MINUS	0x010b		/* none */
#define	K_G_STAR	0x010c		/* none */
#define	K_G_SLASH	0x010d		/* none */

#define	K_RESIZE	0x010e		/* xterm+ncurses */

#define	K_SHIFT_LEFT	(_KS|K_LEFT)	/* aixterm,hft */
#define	K_SHIFT_RIGHT	(_KS|K_RIGHT)	/* aixterm,hft */
#define	K_SHIFT_UP	(_KS|K_UP)	/* aixterm,hft */
#define	K_SHIFT_DOWN	(_KS|K_DOWN)	/* aixterm,hft */
#define	K_SHIFT_HOME	(_KS|K_HOME)	/* aixterm,hft */
#define	K_SHIFT_END	(_KS|K_END)	/* aixterm,hft */
#define	K_SHIFT_PGUP	(_KS|K_PGUP)	/* aixterm,hft */
#define	K_SHIFT_PGDN	(_KS|K_PGDN)	/* aixterm,hft */
#define	K_SHIFT_G_PLUS	(_KS|K_G_PLUS)	/* none */
#define	K_SHIFT_G_MINUS	(_KS|K_G_MINUS)	/* none */
#define	K_SHIFT_G_STAR	(_KS|K_G_STAR)	/* none */
#define	K_SHIFT_G_SLASH	(_KS|K_G_SLASH)	/* none */

#define	K_CTRL_LEFT	(_KC|K_LEFT)	/* aixterm,hft */
#define	K_CTRL_RIGHT	(_KC|K_RIGHT)	/* aixterm,hft */
#define	K_CTRL_UP	(_KC|K_UP)	/* aixterm,hft */
#define	K_CTRL_DOWN	(_KC|K_DOWN)	/* aixterm,hft */
#define	K_CTRL_HOME	(_KC|K_HOME)	/* aixterm,hft */
#define	K_CTRL_END	(_KC|K_END)	/* aixterm,hft */
#define	K_CTRL_PGUP	(_KC|K_PGUP)	/* aixterm,hft */
#define	K_CTRL_PGDN	(_KC|K_PGDN)	/* aixterm,hft */
#define	K_CTRL_INSERT	(_KC|K_INSERT)	/* aixterm,hft */
#define	K_CTRL_DELETE	(_KC|K_DELETE)	/* aixterm,hft */
#define	K_CTRL_G_PLUS	(_KC|K_G_PLUS)	/* aixterm,hft */
#define	K_CTRL_G_MINUS	(_KC|K_G_MINUS)	/* aixterm,hft */
#define	K_CTRL_G_STAR	(_KC|K_G_STAR)	/* aixterm,hft */
#define	K_CTRL_G_SLASH	(_KC|K_G_SLASH)	/* aixterm,hft */

#define	K_ALT_LEFT	(_KA|K_LEFT)	/* aixterm,hft */
#define	K_ALT_RIGHT	(_KA|K_RIGHT)	/* aixterm,hft */
#define	K_ALT_UP	(_KA|K_UP)	/* aixterm,hft */
#define	K_ALT_DOWN	(_KA|K_DOWN)	/* aixterm,hft */
#define	K_ALT_HOME	(_KA|K_HOME)	/* aixterm,hft */
#define	K_ALT_END	(_KA|K_END)	/* aixterm,hft */
#define	K_ALT_PGUP	(_KA|K_PGUP)	/* aixterm,hft */
#define	K_ALT_PGDN	(_KA|K_PGDN)	/* aixterm,hft */
#define	K_ALT_INSERT	(_KA|K_INSERT)	/* aixterm,hft */
#define	K_ALT_DELETE	(_KA|K_DELETE)	/* aixterm,hft */
#define	K_ALT_G_PLUS	(_KA|K_G_PLUS)	/* aixterm,hft */
#define	K_ALT_G_MINUS	(_KA|K_G_MINUS)	/* aixterm,hft */
#define	K_ALT_G_STAR	(_KA|K_G_STAR)	/* aixterm,hft */
#define	K_ALT_G_SLASH	(_KA|K_G_SLASH)	/* aixterm,hft */

#define	K_FN(n)		(0x0200+(n))
#define	K_F1		K_FN(1)		/* Any terminal */
#define	K_F2		K_FN(2)		/* Any terminal */
#define	K_F3		K_FN(3)		/* Any terminal */
#define	K_F4		K_FN(4)		/* Any terminal */
#define	K_F5		K_FN(5)		/* Any terminal */
#define	K_F6		K_FN(6)		/* Any terminal */
#define	K_F7		K_FN(7)		/* Any terminal */
#define	K_F8		K_FN(8)		/* Any terminal */
#define	K_F9		K_FN(9)		/* Any terminal */
#define	K_F10		K_FN(10)	/* Any terminal */
#define	K_F11		K_FN(11)	/* xterm,aixterm,hft,linux */
#define	K_F12		K_FN(12)	/* xterm,aixterm,hft,linux */

#define	K_SHIFT_FN(n)	(_KS|K_FN(n))
#define	K_SHIFT_F1	K_SHIFT_FN(1)	/* aixterm,hft,xenix */
#define	K_SHIFT_F2	K_SHIFT_FN(2)	/* aixterm,hft,xenix */
#define	K_SHIFT_F3	K_SHIFT_FN(3)	/* aixterm,hft,xenix */
#define	K_SHIFT_F4	K_SHIFT_FN(4)	/* aixterm,hft,xenix */
#define	K_SHIFT_F5	K_SHIFT_FN(5)	/* aixterm,hft,xenix */
#define	K_SHIFT_F6	K_SHIFT_FN(6)	/* aixterm,hft,xenix */
#define	K_SHIFT_F7	K_SHIFT_FN(7)	/* aixterm,hft,xenix */
#define	K_SHIFT_F8	K_SHIFT_FN(8)	/* aixterm,hft,xenix */
#define	K_SHIFT_F9	K_SHIFT_FN(9)	/* aixterm,hft,xenix */
#define	K_SHIFT_F10	K_SHIFT_FN(10)	/* aixterm,hft,xenix */
#define	K_SHIFT_F11	K_SHIFT_FN(11)	/* aixterm,hft */
#define	K_SHIFT_F12	K_SHIFT_FN(12)	/* aixterm,hft */

#define	K_CTRL_FN(n)	(_KC|K_FN(n))
#define	K_CTRL_F1	K_CTRL_FN(1)	/* aixterm,hft,xenix */
#define	K_CTRL_F2	K_CTRL_FN(2)	/* aixterm,hft,xenix */
#define	K_CTRL_F3	K_CTRL_FN(3)	/* aixterm,hft,xenix */
#define	K_CTRL_F4	K_CTRL_FN(4)	/* aixterm,hft,xenix */
#define	K_CTRL_F5	K_CTRL_FN(5)	/* aixterm,hft,xenix */
#define	K_CTRL_F6	K_CTRL_FN(6)	/* aixterm,hft,xenix */
#define	K_CTRL_F7	K_CTRL_FN(7)	/* aixterm,hft,xenix */
#define	K_CTRL_F8	K_CTRL_FN(8)	/* aixterm,hft,xenix */
#define	K_CTRL_F9	K_CTRL_FN(9)	/* aixterm,hft,xenix */
#define	K_CTRL_F10	K_CTRL_FN(10)	/* aixterm,hft,xenix */
#define	K_CTRL_F11	K_CTRL_FN(11)	/* aixterm,hft */
#define	K_CTRL_F12	K_CTRL_FN(12)	/* aixterm,hft */

#define	K_ALT_FN(n)	(_KA|K_FN(n))
#define	K_ALT_F1	K_ALT_FN(1)	/* aixterm,hft,xenix */
#define	K_ALT_F2	K_ALT_FN(2)	/* aixterm,hft,xenix */
#define	K_ALT_F3	K_ALT_FN(3)	/* aixterm,hft,xenix */
#define	K_ALT_F4	K_ALT_FN(4)	/* aixterm,hft,xenix */
#define	K_ALT_F5	K_ALT_FN(5)	/* aixterm,hft,xenix */
#define	K_ALT_F6	K_ALT_FN(6)	/* aixterm,hft,xenix */
#define	K_ALT_F7	K_ALT_FN(7)	/* aixterm,hft,xenix */
#define	K_ALT_F8	K_ALT_FN(8)	/* aixterm,hft,xenix */
#define	K_ALT_F9	K_ALT_FN(9)	/* aixterm,hft,xenix */
#define	K_ALT_F10	K_ALT_FN(10)	/* aixterm,hft,xenix */
#define	K_ALT_F11	K_ALT_FN(11)	/* aixterm,hft */
#define	K_ALT_F12	K_ALT_FN(12)	/* aixterm,hft */

#define	K_AN(n)		(_KA|(n))
#define	K_ALT_0		K_AN(0)		/* aixterm,hft,linux */
#define	K_ALT_1		K_AN(1)		/* aixterm,hft,linux */
#define	K_ALT_2		K_AN(2)		/* aixterm,hft,linux */
#define	K_ALT_3		K_AN(3)		/* aixterm,hft,linux */
#define	K_ALT_4		K_AN(4)		/* aixterm,hft,linux */
#define	K_ALT_5		K_AN(5)		/* aixterm,hft,linux */
#define	K_ALT_6		K_AN(6)		/* aixterm,hft,linux */
#define	K_ALT_7		K_AN(7)		/* aixterm,hft,linux */
#define	K_ALT_8		K_AN(8)		/* aixterm,hft,linux */
#define	K_ALT_9		K_AN(9)		/* aixterm,hft,linux */

#define	K_AC(c)		(_KA|(c))
#define	K_ALT_A		K_AC('A')	/* aixterm,hft,linux */
#define	K_ALT_B		K_AC('B')	/* aixterm,hft,linux */
#define	K_ALT_C		K_AC('C')	/* aixterm,hft,linux */
#define	K_ALT_D		K_AC('D')	/* aixterm,hft,linux */
#define	K_ALT_E		K_AC('E')	/* aixterm,hft,linux */
#define	K_ALT_F		K_AC('F')	/* aixterm,hft,linux */
#define	K_ALT_G		K_AC('G')	/* aixterm,hft,linux */
#define	K_ALT_H		K_AC('H')	/* aixterm,hft,linux */
#define	K_ALT_I		K_AC('I')	/* aixterm,hft,linux */
#define	K_ALT_J		K_AC('J')	/* aixterm,hft,linux */
#define	K_ALT_K		K_AC('K')	/* aixterm,hft,linux */
#define	K_ALT_L		K_AC('L')	/* aixterm,hft,linux */
#define	K_ALT_M		K_AC('M')	/* aixterm,hft,linux */
#define	K_ALT_N		K_AC('N')	/* aixterm,hft,linux */
#define	K_ALT_O		K_AC('O')	/* aixterm,hft,linux */
#define	K_ALT_P		K_AC('P')	/* aixterm,hft,linux */
#define	K_ALT_Q		K_AC('Q')	/* aixterm,hft,linux */
#define	K_ALT_R		K_AC('R')	/* aixterm,hft,linux */
#define	K_ALT_S		K_AC('S')	/* aixterm,hft,linux */
#define	K_ALT_T		K_AC('T')	/* aixterm,hft,linux */
#define	K_ALT_U		K_AC('U')	/* aixterm,hft,linux */
#define	K_ALT_V		K_AC('V')	/* aixterm,hft,linux */
#define	K_ALT_W		K_AC('W')	/* aixterm,hft,linux */
#define	K_ALT_X		K_AC('X')	/* aixterm,hft,linux */
#define	K_ALT_Y		K_AC('Y')	/* aixterm,hft,linux */
#define	K_ALT_Z		K_AC('Z')	/* aixterm,hft,linux */

#ifdef AIX_EXTENDED_CURSES
#define	TH_SAVE_RESTORE_CELL 6
#endif

#ifndef TH_C

extern int x_size, y_size;

extern TH_BOOLEAN th_init(int mode, int x_size, int y_size);
extern void th_deinit(void);
extern void th_beep(void);
extern void th_cls(void);
extern void th_cxy(int x, int y);
extern void th_cy(int y);
extern void th_chome(void);
extern void th_setcol(int col);
extern void th_setcsr(int csr);
extern void th_inslin(void);
extern void th_dellin(void);
extern void th_deleol(void);
#ifdef AIX_EXTENDED_CURSES
extern void th_save(int x, int y, int n, void *buf);
extern void th_restore(int x, int y, int n, void *buf);
#endif
extern TH_BOOLEAN th_isprint(int c);
extern void th_p_raw_chr(int c);
extern void th_p_raw_str(const char *s);
extern void th_p_chr(int c);
extern void th_p_str(const char *s);
extern void th_sync_screen(void);

extern int th_key_read(void);
#if defined(AIX) || defined(LINUX) || defined(HP) || defined(SUN) || defined(S390USS) || defined(CYGWIN) || defined(MACOSX) || defined(IPHONE)
extern int th_key_status(void);
#endif
extern TH_BOOLEAN th_key_break(void);

#endif

#endif
